@@ -44,15 +44,24 @@ def token_to_str(token: str):
     }.get(token, "'" + token.lower() + "'")
 
 class File():
-    def __init__(self, parent=None):
+    def __init__(self, parent=None, root=None):
         self.parent = parent
+        self.root = root or self
         self.settings = {}
         self.dependencies = []
         self.things = []
+        self.diagnostics = []
+        if self.parent is None:
+            self.all_loaded = set()
 
-    def load_from_text(self, source):
+    def load_from_text(self, source, path=None):
         self.source = source
-        self.path = "<from source>"
+        self.path = path.abspath(path) or "<from source>"
+
+        if self.parent is None:
+            self.all_loaded.add(self.path)
+
+        log.verbose(f"Loaded from source: {self.path} {Fore.LIGHTBLACK_EX}{'(root)' if not self.parent else ''}", "load")
 
     def load_from_file(self, source: str|TextIOWrapper):
         # read the file
@@ -60,14 +69,18 @@ class File():
             self.path = source
             source = open(source, "r")
         else:
-            self.path = source.name
+            self.path = path.abspath(source.name)
         self.source = source.read()
         source.close()
 
-        log.verbose(f"Loaded file: {Fore.WHITE}{self.path}{Fore.LIGHTBLACK_EX} {'(root)' if not self.parent else ''}")
+        if self.parent is None:
+            self.all_loaded.add(self.path)
+
+        log.verbose(f"Loaded file: {self.path} {Fore.LIGHTBLACK_EX}{'(root)' if not self.parent else ''}", "load")
 
     def search_paths(self):
         return [
+            "",
             ".",
             path.dirname(self.path), # next to this file
             path.join(path.dirname(__file__), "stdlib") # in the standard library
@@ -104,71 +117,135 @@ class File():
         except UnexpectedInput as e:
             return e.expected, e.state.value_stack
 
-        return None # no expected tokens here
+        return None, None # no expected tokens here
+
+    def __parsing_error(self, e: lark.UnexpectedToken):
+        log.verbose(f"Parsing error: {e}", "corrector")
+
+        parser = e.interactive_parser
+        tok: Token = e.token
+        token = tok.value.split(' ')[0]
+        expected = ', '.join(token_to_str(t) for t in e.expected)
+        error_text = f"Expected{' one of:' if len(e.expected) > 1 else ''} {expected}"
+        dur = len(token)
+
+        if token == "":
+            # empty token = EOF
+            line = self.source.split('\n')[e.line - 1]
+            location = Location(self, e.line, len(line) + 1, 0)
+        else:
+            location = Location(self, e.line, e.column, dur)
+
+        diag = Diagnostic([location], DiagLevel.ERROR, error_text)
+        self.diagnostics += [diag]
+
+        # inform the user about our naming conventions :)
+        # while trying to rename
+        inter = e.expected.intersection({"TYPE_IDENTIFIER", "ROOT_IDENTIFIER"})
+        if len(inter):
+            diag.message = "This identifier should use PascalCase"
+            tok.type = inter.pop()
+            tok.value = tok.value[0].upper() + tok.value[1:]
+            log.verbose(f"Renamed '{token}' to '{tok.value}'", "corrector")
+            parser.feed_token(tok)
+            return True
+
+        inter = e.expected.intersection({"FIELD_IDENTIFIER", "METHOD_IDENTIFIER", "VALIDATOR_IDENTIFIER"})
+        if len(inter):
+            diag.message = "This identifier should use snake_case"
+            tok.type = inter.pop()
+            tok.value = tok.value.lower()
+            parser.feed_token(tok)
+            return True
+
+        # fill in missing semicolons and things
+        fill_in = {
+            "SEMICOLON": ";", "COLON": ":",
+            "RPAR": ")", "RBRACE": "}", "RSQB": "]",
+            "COMMA": ",",
+        }
+        for k, v in fill_in.items():
+            if k in e.expected:
+                log.verbose(f"Inserted {k}", "corrector")
+                parser.feed_token(Token(k, v))
+
+                # insert an additional semicolon if the next token is not it
+                if k in {"RPAR", "RSQB"} and tok.type != "SEMICOLON":
+                    parser.feed_token(Token("SEMICOLON", ";"))
+                    log.verbose(f"Inserted SEMICOLON", "corrector")
+
+                if tok.type == "ROOT_IDENTIFIER":
+                    tok.type = "TYPE_IDENTIFIER"
+                log.verbose(f"Inserted original token", "corrector")
+                print(parser.parser_state.value_stack, tok.type, tok.value)
+                parser.feed_token(tok)
+
+                return True
+
+        # fill in missing numeric values
+        structure = ([None, None] + parser.parser_state.value_stack)[-2]
+        if e.expected == {"LPAR"} and structure and structure.type in {"ENTITY", "GLOBALMETHOD", "METHOD", "STATICMETHOD", "CONFIRMATION"}:
+            diag.message = "Missing numeric value"
+            log.verbose(f"Inserted '(0)'", "corrector")
+            parser.feed_token(Token("LPAR", "("))
+            parser.feed_token(Token("NUMBER", "0"))
+            parser.feed_token(Token("RPAR", ")"))
+            parser.feed_token(tok)
+            return True
+
+        return False
 
     def parse(self) -> Tuple[list[SusThing], list[Diagnostic]]:
-        log.verbose(f"Parsing {Fore.WHITE}{self.path}")
+        log.verbose(f"Parsing {Fore.WHITE}{self.path}", "parser")
         self.things = []
         self.dependencies = []
-        diag = []
+        self.diagnostics = []
 
         try:
-            self.tree = lark_parser.parse(self.source)
-            log.verbose(f"AST constructed")
+            self.tree = lark_parser.parse(self.source, on_error=self.__parsing_error)
+            log.verbose(f"AST constructed", "parser")
         except UnexpectedInput as e:
-            token = e.token.value.split(' ')[0]
-            expected = ', '.join(token_to_str(t) for t in e.expected)
-            error_text = f"Unexpected input. Expected{' one of:' if len(e.expected) > 1 else ''} {expected}"
-            dur = len(token)
-
-            if token == "":
-                # empty token = EOF
-                line = self.source.split('\n')[e.line - 1]
-                location = Location(self, e.line, len(line) + 1, 0)
-            else:
-                location = Location(self, e.line, e.column, dur)
-
-            diag = Diagnostic([location], DiagLevel.ERROR, error_text)
-
-            # inform the user about our naming conventions :)
-            if len(e.expected.intersection({"TYPE_IDENTIFIER", "ROOT_IDENTIFIER"})):
-                diag.message += "\nHint: identifiers for things except methods, fields and members use PascalCase"
-            if len(e.expected.intersection({"FIELD_IDENTIFIER", "METHOD_IDENTIFIER", "VALIDATOR_IDENTIFIER"})):
-                diag.message += "\nHint: method, field, member and validator identifiers use snake_case"
-
+            log.verbose("No corrections or invalid correction", "corrector")
             # parsing can't continue any further, just return
-            return [], [diag]
+            return [], self.diagnostics
 
         # deconstruct the syntax tree
         for thing in self.tree.children:
             if thing.data == "inclusion":
                 name = thing.children[0]
-                log.verbose(f"Encountered inclusion:{Fore.LIGHTBLACK_EX} path={Fore.WHITE}{name}")
+                log.verbose(f"Encountered inclusion:{Fore.LIGHTBLACK_EX} path={Fore.WHITE}{name}", "parser")
                 # find dependency
                 try:
                     source = self.resolve_source(name.value)
                 except SearchError as ex:
                     return [], [Diagnostic([Location(self, name.line, name.column, len(name))],
                         DiagLevel.ERROR, ex.msg)]
+
                 # load it
-                dependency = File(self)
-                dependency.load_from_file(source)
-                self.dependencies.append(dependency)
+                if source.name not in self.root.all_loaded:
+                    dependency = File(self, self.root)
+                    dependency.load_from_file(source)
+                    self.dependencies.append(dependency)
+                    self.root.all_loaded.add(source.name)
+                else:
+                    return [], [Diagnostic([Location(self, name.line, name.column, len(name))],
+                        DiagLevel.ERROR, "This file has already been included directly or by a dependency within this project\n" +\
+                        f"Note: inclusion resolved to '{source.name}'")]
 
             elif thing.data == "setting":
                 name = thing.children[0]
                 value = thing.children[1]
-                log.verbose(f"Encountered setting:{Fore.LIGHTBLACK_EX} name={Fore.WHITE}{name}{Fore.LIGHTBLACK_EX} value={Fore.WHITE}{value}")
+                log.verbose(f"Encountered setting:{Fore.LIGHTBLACK_EX} name={Fore.WHITE}{name}{Fore.LIGHTBLACK_EX} value={Fore.WHITE}{value}", "parser")
                 if name.value not in KNOWN_SETTINGS:
-                    diag.append(Diagnostic([Location(self, name.line, name.column, len(name))], DiagLevel.WARN, "Unknown setting"))
+                    self.diagnostics.append(Diagnostic([Location(self, name.line, name.column, len(name))], DiagLevel.WARN, "Unknown setting"))
                 self.settings[name.value] = value.value
 
             else:
                 if thing.data == "definition":
                     thing = thing.children[0]
-                log.verbose(f"AST subtree: {log.highlight_ast(thing)}")
+                log.verbose(f"AST subtree: {log.highlight_ast(thing)}", "parser")
                 thing = convert_ast(thing, self)
-                log.verbose(f"Converted AST subtree: {Fore.WHITE}{log.highlight_thing(thing)}")
+                log.verbose(f"Converted AST subtree: {Fore.WHITE}{log.highlight_thing(thing)}", "parser")
                 self.things.append(thing)
                 # generate standard methods for entities
                 if isinstance(thing, SusEntity):
@@ -197,18 +274,20 @@ class File():
                         None
                     ))
 
+        log.verbose(f"Parsing dependencies for {Fore.WHITE}{self.path}", "deps")
         # parse dependencies
         things = self.things
         for dep in self.dependencies:
             things += dep.parse()[0]
+        log.verbose(f"Parsing dependencies done for {Fore.WHITE}{self.path}", "deps")
 
         # run linker
-        diag = []
-        if self.parent == None:
+        if not self.parent:
             things, diag = linker.run(things)
+            self.diagnostics += diag
 
         self.things = things
-        return things, diag
+        return things, self.diagnostics
 
     def write_output(self, lang, target_dir):
         if not self.things:
